#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Mar  2 18:22:06 2020

@author: fabian

This module contains a class that makes the alignments and superimpositions of 
the chains in order to assembly them. However, it does not join the chains, as some 
extra conditions may be required to do so. For example, add only necessary chain
for a desire stoichiometry
"""

import copy

from Bio import PDB
from Bio import pairwise2

class Ensemble():
    
    io = PDB.PDBIO()
    
    '''
        This class takes structures of proteins (Biopyton PDBs) and superimposes them
        in order to make one bigger structure with all the chains given
    '''
    
    def __init__(self, structure_1, structure_2):
        '''
            
        '''
        self.structure_A, self.structure_B = structure_1, structure_2
        self.superimposer = PDB.Superimposer()
        self._rms = None
    
    def get_best_alignment(self, fasta_seqs):
        '''
            Obtains the best alignment between 2 fasta files sequences.
            
            The best alignment shows which chain is common between those files. 
            To know which two sequences are more similar, we make an alignment to determine if they are homologus
            
            Returns:
                - The alignment with best score, which is supposed to be the repeated sequence

        '''
        
        aligned = []
        alignments = []
        # Each possible pair of chains is aligned and all of them are returned
        for chain_A in fasta_seqs[0]:
            seq1 = fasta_seqs[0][chain_A]
            for chain_B in fasta_seqs[1]:
                seq2 = fasta_seqs[1][chain_B]
                #if (seq1 == seq2): continue We make all posibilities so we don't skip e
                if (chain_B, chain_A) in aligned: continue
                aligned.append((chain_A, chain_B))
                # The alignment algorithms does several of them, so we take only the best one
                aligned_pair = pairwise2.align.globalxx(seq1, seq2)
                max_score = max(aligned_pair, key=lambda x: x[2])
                alignments.append(((chain_A, chain_B), max_score[2]/len(max_score[0])))
        # We return the chain ids of the best alignment and its score
        return alignments
    
    def superimpose(self, chain_A, chain_B):
        '''
            Superimpose two complexes given two chains. This chains should be homologous, 
            and they are used to calculate the matrices and rotations needed to be applied 
            to the rest of the chains
            
            input: The chain id to superimpose of structure A and the chain to superimpose of chain B
            returns a list with the atoms with their coordinates changed
        '''
        # Obtaining the chains
        chain_A = next(filter(lambda x: x.id == chain_A, self.structure_A.get_chains()))
        chain_B = next(filter(lambda x: x.id == chain_B, self.structure_B.get_chains()))
        # Obtaining the chains atoms
        moving_chains = copy.copy(self.structure_B)
        model = next(moving_chains.get_models())
        moving_chains = model.get_chains()
        superimposed_atoms = list(chain_B.get_atoms())
        chain_atoms = list(chain_A.get_atoms())
        # If both sequences have different number of atoms, get only the common part
        # This can be improved using alignments
        if (len(list(chain_A.get_atoms())) != len(list(chain_B.get_atoms()))):
            length = min(len(list(chain_A)), len(list(chain_B)))
            chain_atoms = chain_atoms[:length]
            superimposed_atoms = superimposed_atoms[:length]
        # Doing the superimpositions and returning a list with the atoms
        # With the changed positions
        self.superimposer.set_atoms(chain_atoms, superimposed_atoms)
        chains_changed = list()
        for chain in moving_chains:
            if chain.id == chain_B.id: continue # We skip the homologous one, it's present in both structures
            atoms = chain.get_atoms()
            self.superimposer.apply(atoms)
            chains_changed.append(chain)
        self._rms = self.superimposer.rms
        return chains_changed
    
    @property
    def chains(self):
        return self.structure_A, self.structure_B
       
    @property
    def rms(self):
        '''
            Returns the root-mean-square distance of the superimposed proteins.
            If superimposition wasn't done raises an error.
        '''
        if(self._rms is None):
            raise RuntimeError('Cannot determine rms unless superimposition is done')
        else:
            return self._rms
    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Created on Tue Feb 25 19:39:13 2020

@author: fabian

@version: 1.0

Splits a PDB into its component chains, each in one file. If indicated, 
it will also create fasta files for each chain

"""

from Bio import PDB

import os, os.path
import argparse
import sys

AA = {'ALA':'A', 'ARG':'R', 'ASN':'N', 'ASP':'D',
      'CYS':'C', 'GLN':'Q', 'GLU':'E', 'GLY':'G',
      'HIS':'H','ILE':'I','LEU':'L','LYS':'K',
      'MET':'M', 'PHE':'F','PRO':'P','SER':'S',
      'THR':'T', 'TRP':'W','VAL':'V','TYR':'Y', 'TERM':'',
      'UNK':'X',' DA':'A',' DG':' G', ' DC':'C', 
      ' DT':'T','  U':'U','  A':'A','  G':'G', '  C':'C'}

def _parse_args():
    '''
        Parses the arguments needed for this script to work
    '''
    parser = argparse.ArgumentParser()
    parser.add_argument('-i','--input', required=True, dest='input', action='store')
    parser.add_argument('-p','--prefix', required=True, dest='prefix', action='store')
    parser.add_argument('-o','--output', required=True, dest='output', action='store')
    parser.add_argument('-f','--fasta', action='store_true', dest='fasta', default=False)
    return parser.parse_args()


if(__name__ == '__main__'):
    '''
        Main routine of the program
        
        Checks the arguments to know if the input is correct. Then tries to parse the pdb structure indicated in arguments.input
        (may throw a Bio.PDB warning though). If output folder is a file, exits with error -1. If not, create the folder in case it doesn't exists and continues.
        
        For every chain of the complex, creates a new empty structure object, adds the chain and saves it with the estructure {prefix}_{chain_id}.pdb.
        In case the flag -f was indicate, also saves the fasta sequence in {prefix}_{chain_id}.fa. Both files are saved in the output directory
    '''
    arguments = _parse_args()
    if(os.path.isfile(arguments.input)):  # If the input is a file, tries to read it
        pdb_file = arguments.input
        PDBparser = PDB.PDBParser(QUIET=True)
        structure = PDBparser.get_structure(pdb_file[:-3], pdb_file) # Parses the PDB
        io = PDB.PDBIO()
        if(os.path.exists(arguments.output) and os.path.isfile(arguments.output)): #Checks the output is correct and either doesn't exists or it's a folder
            print('File {out} already exists!'.format(out=arguments.output), file=sys.stderr)
            sys.exit(-1) # Exits because output is a file and already exists
        elif(not os.path.exists(arguments.output)): # Folder does not exist, so let's create it
            os.mkdir(arguments.output)
        # For each chain save it in a separate file. Log is provided for very long proteins
        for chain in structure.get_chains():
            splitted = PDB.Structure.Structure(pdb_file[:-3]+chain.id)
            splitted.add(chain)
            io.set_structure(chain)
            print('Saving chain {chain_id} in {prefix}_{chain_id}.pdb'.format(prefix=arguments.prefix, chain_id=chain.id))
            # Saves also the fasta file if indicated with -f flag
            if (arguments.fasta):
                print('Saving sequence of chain {chain_id} in {prefix}_{chain_id}.fa'.format(prefix=arguments.prefix, chain_id=chain.id))
                filename = os.path.join(arguments.output, '{prefix}_{chain_id}.fa'.format(prefix=arguments.prefix, chain_id=chain.id))
                fhand = open(filename, 'w')
                fhand.write('>{prefix}_{chain_id}\n'.format(prefix=arguments.prefix, chain_id=chain.id))
                for residue in chain:
                    fhand.write(AA.get(residue.resname, '').replace(' ',''))
                fhand.write('\n')
                fhand.close()
            filename = os.path.join(arguments.output, '{prefix}_{chain_id}.pdb'.format(prefix=arguments.prefix, chain_id=chain.id))
            io.save(filename)
    elif(not os.path.exists(arguments.input)): # If there's no input file
        print('File {file} does not exist'.format(file=arguments.input), 
              file=sys.stderr)
    else: # User gave a folder as input
        print('Unknown input file or is not a file', 
              file=sys.stderr)
        


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Mar  2 12:11:19 2020

@author: fabian

@version: 0.9

Joins all the PDBs in a folder into pairs, but only if they interact
"""

from Bio import PDB
import os
import os.path
import sys
import re

def join_fasta(*fastas):
    text = ''
    for fhand in fastas:
        with open(fhand) as f1:
            for line in f1:
                text += line 
    return text

if __name__ == '__main__':
    directory = sys.argv[1] # Folder with the chain pdbs
    distance = float(sys.argv[2])
    directory_out = os.path.join(directory, 'pairs') # Folder where pairs will be saved
    PDBparser = PDB.PDBParser(QUIET=True)
    done = []
    for pdb_1 in list(filter(lambda x: x.endswith('.pdb'), os.listdir(directory))):
        for pdb_2 in list(filter(lambda x: x.endswith('.pdb'), os.listdir(directory))):
            if (pdb_1 == pdb_2):   
                continue # Not to duplicate itself
            elif not (pdb_1, pdb_2) in done or not (pdb_2, pdb_1) in done: # Not to duplicate pairs
                done.append((pdb_1, pdb_2))
                structure1 = PDBparser.get_structure(pdb_1[:-4], os.path.join(directory,pdb_1))
                structure2 = PDBparser.get_structure(pdb_2[:-4], os.path.join(directory,pdb_2))
                for chain in structure2.get_chains():
                    atoms = list(chain.get_atoms()) 
                    ns = PDB.NeighborSearch(atoms) # An object to search chains near an atom 
                    for target_atom in structure1.get_atoms():
                        near = ns.search(target_atom.coord, distance)
                        if(near and (pdb_1[-5:-4],pdb_2[-5:-4]) not in done and (pdb_2[-5:-4],pdb_1[-5:-4]) not in done): # If there's an atom near, they interact
                            print('Joining chain {a} and {b}'.format(a=pdb_1[-5:-4], b=pdb_2[-5:-4]))
                            done.append((pdb_1[-5:-4],pdb_2[-5:-4]))
                            io = PDB.PDBIO()
                            structure1[0].child_list += [chain for chain in structure2.get_chains()] # Joins all the chains
                            io.set_structure(structure1)
                            if(not os.path.exists(os.path.join(directory_out))):
                                os.mkdir(os.path.join(directory_out))
                            p = re.search('^[A-Za-z0-9]+_([A-Za-z0-9]+).pdb$', pdb_1 ).group(1)
                            q = re.search('^[A-Za-z0-9]+_([A-Za-z0-9]+).pdb$', pdb_2 ).group(1)
                            io.save(os.path.join(directory_out,pdb_1[:-5]+p+q+'.pdb'))
                            print(os.path.join(directory,pdb_1[:-4]+'.fa'))
                            if(os.path.exists(os.path.join(directory,pdb_1[:-4]+'.fa'))) and os.path.exists(os.path.join(directory,pdb_2[:-4]+'.fa')):
                                path = os.path.join(directory_out,pdb_1[:-5]+p+q+'.fa')
                                file = open(path, 'w')
                                file.write(join_fasta(os.path.join(directory,pdb_1[:-4]+'.fa'), os.path.join(directory,pdb_2[:-4]+'.fa')))
                                file.close()



# -*- coding: utf-8 -*-

"""
Created on Tue Feb 25 19:39:13 2020

@author: fabian

@version: 1.0

"""

from .superimpose import Ensemble
from . import errors

from Bio import PDB
from Bio import pairwise2

def _chain_id():
    '''
        The differents id the chains will have inside the model. Now supports a model up to 60 different ids:
        25 for uppercase letters, 25 for lowercase letters and 10 more from numbers 0 to 9.
    '''
    for i in range(65, 91, 1):
        yield chr(i)
    for i in range(97, 123, 1):
        yield i
    for i in range(48, 58, 1):
        yield i

AA_EQUIVALENCE = {'ALA':'A', 'ARG':'R', 'ASN':'N', 'ASP':'D',
      'CYS':'C', 'GLN':'Q', 'GLU':'E', 'GLY':'G',
      'HIS':'H','ILE':'I','LEU':'L','LYS':'K',
      'MET':'M', 'PHE':'F','PRO':'P','SER':'S',
      'THR':'T', 'TRP':'W','VAL':'V', 'TYR':'Y','TERM':'',
      'UNK':'X','DA':'A','DG':'G', 'DC':'C', 
      'DT':'T','U':'U','A':'A','G':'G', 'C':'C'}

def _get_chain_sequence(chain):
    '''
        Obtains the total number of different chains that are in the different structures
    '''
    chain_sequence = ''
    for residue in chain.get_residues():
        residue_val = residue.resname.lstrip().strip()
        chain_sequence += AA_EQUIVALENCE.get(residue_val,'')
    return chain_sequence            
    

def get_fastas_from_structs(structs, fastas, threshold=0.9):
    '''
        Determines which of the sequences of the fastas correspond to each of the pdb chains
        Two similar chains can refer to the same fasta sequence. 
        
        It will raise an exception if a schain in the pdb does not have an homologous in the fasta. 
    '''
    chain_fasta = {}
    for struct in structs:
        chain_fasta[struct.id] = {}
        for chain in struct.get_chains():
            chain_sequence = _get_chain_sequence(chain)
            #print(chain_sequence)
            for seq in fastas:
                #print(seq.seq)
                alignment = pairwise2.align.globalxx(chain_sequence, seq.seq, one_alignment_only=True)
                score = alignment[0][2]/len(alignment[0][0])
                if score > threshold:
                    chain_fasta[struct.id][chain.id] = seq
                    break
            else:
                raise errors.PDB_disagrees_fasta(struct.id)
                pass
    return chain_fasta

def build_complex(threshold, distance, stoichiometry, sequences, structures, verbose, initial):
    '''
        This is the core function of the program. 
        
        Takes a list of structures and tries to join them in a single model, taking into account
        diferent parameters, such as an intial model, a threshold for the pairwise alignments and 
        the distance to consider that two chains are in the same position and it must be discarded 
        (different atoms cannot occupy the same space)
    '''
    # Initiate empty structure and check some errors
    seqs = get_fastas_from_structs(structures, sequences)
    
    # Initializing an empty structure to save the model
    full_structure = PDB.Structure.Structure('full')
    full_structure.add(PDB.Model.Model('model'))
    # We will need to assign ids. For now is limited to up to 64 different chains, but more single characters can be added
    ids = _chain_id()
    iters = 0
    pair_num = len(structures)
    # Prepare a dict that saves how many chains of every one in the stoichiometry has the model in construction
    if(stoichiometry):
        current_number_of_chains = {chain_id:0 for chain_id in set(stoichiometry)}
    # We need at least 2 pdbs with an interaction. Can be the same one repeated. This is done to avoid trouble
    if(pair_num < 2):
        raise ValueError('Needed at least 2 pdbs to superpose')
    failed = []
    current = 0
    done = []
    
    
    # If an initial structure was given, introduce it into the model
    if(initial):
        if verbose: print('Initializing complex')
        # Obtaining the sequences of the chains
        init_seqs = get_fastas_from_structs([initial], sequences)
        seqs[full_structure.id] = {}
        # For each chain, add it to the model with sequential ids: A->B-> etc
        for idx, chain in enumerate(initial.get_chains()):
            chain_id = next(ids)
            seqs[full_structure.id][chain_id] = init_seqs[initial.id][chain.id]
            chain_new = PDB.Chain.Chain(chain_id)
            chain_new.child_list = list(chain.get_residues())
            model = next(full_structure.get_models())
            # If stoichiometry is selected, then count which sequences will be added
            if(stoichiometry):
                chain_sequence = _get_chain_sequence(chain_new)
                for seq in sequences:
                    alignment = pairwise2.align.globalxx(chain_sequence, seq.seq, one_alignment_only=True) 
                    # This chain corresponds to this stoichiometry chain
                    if(alignment[0][2]/len(alignment[0][0])> threshold and
                       current_number_of_chains[seq.id] < stoichiometry[seq.id]):
                        model.child_list.append(chain_new)
                        current_number_of_chains[seq.id] += 1
                        if(stoichiometry):
                            count = 0
                            for chain in current_number_of_chains:
                                if current_number_of_chains[chain] == stoichiometry[chain]:
                                    count += 1
                                if count == len(stoichiometry):
                                    print('Stoichiometry fullfilled!')
                                    return full_structure
            model.child_list.append(chain_new)
        done.append(initial.id)
    if verbose: print('Starting to build')
    # The main loop of the function
    while True:
        if verbose: 
            print('Loop #%i' % current)
            print('Current number of chains: %i' % len(list(full_structure.get_chains())))
        # All the structures where added correctly (should only be for non stoichiometric uses)
        if(not stoichiometry and len(done) == len(structures)): 
            break
        structure = structures[current % len(structures)]
        if(structure.id in done): 
            current += 1
            continue # If is already in, continue to the next one
        count = list(failed.count(element) for element in failed)
        # If failed has an element twice, that means that no more available chains could be added
        # Therefor, it will start an endless loop, as the structure remain equal no matter which 
        # of the left structures is trying to be added. We stop it here
        if 3 in count: # We are repeating structures
            if verbose: print('Some pdbs could not be joined.\nAnd endless loop started and no more chains could be added\nFinishing the model in the current state')
            break
        if len(list(full_structure.get_chains())) == 0:
            if verbose: print('Initializing complex')
            seqs[full_structure.id] = {}
            for idx, chain in enumerate(structure.get_chains()):
                chain_id = next(ids) # Selecting the next id available
                seqs[full_structure.id][chain_id] = seqs[structure.id][chain.id] # Assigning the new chains its sequence
                # Empty new chain with the new id
                chain_new = PDB.Chain.Chain(chain_id)
                # Adding the residues to the new chain and adding the chain to the new model
                chain_new.child_list = list(chain.get_residues())
                model = next(full_structure.get_models())
                # Check stoichiometry if needed
                if(stoichiometry):
                    # Need the exact sequence to align
                    chain_sequence = _get_chain_sequence(chain_new)
                    # Align with every sequence in the fasta to know which sequence is 
                    for seq in sequences:
                        alignment = pairwise2.align.globalxx(chain_sequence, seq.seq, one_alignment_only=True) 
                        try:
                            current_number_of_chains[seq.id]
                        except KeyError:
                            # The sequence chain is not in the stoichiometry
                            raise errors.chain_in_stoic_not_in_fasta(seq.id)
                        # If homologous and does not surpass the stoichiometry, add it
                        if(alignment[0][2]/len(alignment[0][0])> threshold and
                            current_number_of_chains[seq.id] < stoichiometry[seq.id]):
                            model.child_list.append(chain_new)
                            current_number_of_chains[seq.id] += 1
                            count = 0
                            for chain in current_number_of_chains:
                                if current_number_of_chains[chain] == stoichiometry[chain]:
                                    count += 1
                                if count == len(stoichiometry):
                                    print('Stoichiometry fullfilled!')
                                    return full_structure
                model.child_list.append(chain_new)

            done.append(structure.id)
            current += 1
            continue
        # The Enseble class keeps code ordered and abstracts the alignment
        # and superimposition of the code
        pair = Ensemble(full_structure, structure)
        pair_seqs = (seqs[full_structure.id], seqs[structure.id])
        alignment = pair.get_best_alignment(pair_seqs)
        # If score is > 0.95, they should be homologs
        for align in alignment:
            if(align[1] > threshold):
                #try:
                atoms_of_chains = pair.superimpose(align[0][0], align[0][1])
                #except PDB.PDBExceptions.PDBException:
                #    current += 1
                #    continue
                iters += 1
                neighbor = PDB.NeighborSearch(list(full_structure.get_atoms()))
                clashes = 0
                for chain in atoms_of_chains:
                    for atom in atoms_of_chains:
                        for atom in chain.get_atoms():
                            close_atoms = neighbor.search(atom.get_coord(), distance)
                            # If there are atoms within 2 angstroms, consider a clash
                            if len(close_atoms) > 0:
                                clashes += 1
                if (clashes < 10 and pair.rms < 0.05):
                    # The chain can be added, so let's add it
                    for chain in atoms_of_chains:
                        chain_id = next(ids) # Chosing an id
                        chain2 = PDB.Chain.Chain(chain_id) # An empty chain which will add the residues to join
                        chain2.child_list += list(chain.get_residues())
                        model = next(full_structure.get_models())
                        if(stoichiometry): # Taking the stoichiometry into account
                            chain_sequence = _get_chain_sequence(chain)
                            for seq in sequences:
                                alignment = pairwise2.align.globalxx(chain_sequence, seq.seq, one_alignment_only=True) 
                                # This chain corresponds to this stoichiometry chain
                                if(alignment[0][2]/len(alignment[0][0])> threshold and
                                           current_number_of_chains[seq.id] < stoichiometry[seq.id]):
                                    model.child_list.append(chain2)
                                    current_number_of_chains[seq.id] += 1
                                    count = 0
                                    for chain in current_number_of_chains:
                                        if current_number_of_chains[chain] == stoichiometry[chain]:
                                            count += 1
                                    if verbose: print('Current number of chains: %i' % sum(current_number_of_chains.values()))
                                    if count == len(stoichiometry):
                                        if verbose: print('Stoichiometry fullfilled!')
                                        return full_structure
                        else:
                            model.child_list.append(chain2)
                            if verbose and stoichiometry: print('Current number of chains: %i' % sum(current_number_of_chains.values()))
                            done.append(structure.id)
                    failed = []
                    break
        else:
            failed.append(structure.id)
            # Pdbs clash
        current += 1
    return full_structure


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Mar  4 10:32:34 2020

@author: fabian
"""

from distutils.core import setup
from os import path

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'theory.md'), encoding='utf-8') as f:
    long_description_in = f.read()

setup(name='promod',
      version='0.1.5',
      description='A macromolecular complex builder',
      long_description=long_description_in,
      long_description_content_type="text/markdown",
      author='Fabián Robledo, Claudio Díaz',
      author_email='fabian.robledo01@estudiant.upf.edu',
      license='MIT',
      keywords='bioinformatics structural_bioinformatics protein complex builder modeler',
      packages=['builder'],
      scripts=['promod', 'promod-tk','scripts/pairpdbs.py', 'scripts/pdbsplit.py'],
      install_requieres = ['Biopython'],
      url='https://github.com/Fabian-RY/SBI-Python-project',
      classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
      ],
      python_requires='>=3',
      include_package_data=True)